@@ -44,6 +44,7 @@
 #include <windows.h>
 
 #include "pechecksum.h"
+#include "imagehelper/pechecksum_update.h"
 
 #if defined(__MSYS__)
 /* MSYS has no strtoull */
@@ -67,12 +68,12 @@ static WORD coff_characteristics_show;
 static WORD pe_characteristics_set;
 static WORD pe_characteristics_clr;
 static WORD pe_characteristics_show;
-static short checksum_show;
-static short checksum_update; /* 1 = update after changes, 2 = fix always */
+static short checksum_option; /* 1 = show, 2 = verify, 3 = fix */
 
 typedef struct
 {
   const char *pathname;
+  unsigned long long filesize;
   PIMAGE_DOS_HEADER dosheader;
   union
     {
@@ -507,24 +508,48 @@ do_mark (const char *pathname)
 
     }
 
-  /* Any actual changes? */
-  if (pep->is_64bit)
-    {
-      changed = memcmp(&old_ntheader.ntheader64, pep->ntheader64,
-		       sizeof(old_ntheader.ntheader64));
-      hdr_checksum = pep->ntheader64->OptionalHeader.CheckSum;
-    }
-  else
-    {
-      changed = memcmp(&old_ntheader.ntheader32, pep->ntheader64,
-		       sizeof(old_ntheader.ntheader32));
-      hdr_checksum = pep->ntheader32->OptionalHeader.CheckSum;
-    }
+  /* Update the checksum. */
+  changed = FALSE;
+  {
+    const void *p_oldheader, *p_newheader;
+    unsigned headersize;
+    ULONG *p_checksum;
+    if (pep->is_64bit)
+      {
+	p_oldheader = &old_ntheader.ntheader64;
+	p_newheader = pep->ntheader64;
+	headersize = sizeof(old_ntheader.ntheader64);
+	p_checksum = &pep->ntheader64->OptionalHeader.CheckSum;
+      }
+    else
+      {
+	p_oldheader = &old_ntheader.ntheader32;
+	p_newheader = pep->ntheader32;
+	headersize = sizeof(old_ntheader.ntheader32);
+	p_checksum = &pep->ntheader32->OptionalHeader.CheckSum;
+      }
+
+    hdr_checksum = pe32_checksum_update_end(
+      pe32_checksum_update_add(
+	pe32_checksum_update_begin(*p_checksum, pep->filesize),
+	p_oldheader, p_newheader, headersize
+      ),
+      pep->filesize
+    );
+
+    if (!hdr_checksum)
+      fprintf(stderr, "%s: Checksum update failed\n", pathname);
+    else if (*p_checksum != hdr_checksum)
+      {
+	changed = TRUE;
+	*p_checksum = hdr_checksum;
+      }
+  }
 
   pe_close (pep);
 
-  /* Checksum calculation requires re-open as a regular file. */
-  if (checksum_show || checksum_update)
+  /* Checksum re-calculation requires re-open as a regular file. */
+  if (checksum_option)
     do_checksum (pathname, indent, changed, hdr_checksum);
   return 0;
 }
@@ -625,16 +650,16 @@ static void
 do_checksum (const char *pathname, int indent, int changed, ULONG hdr_checksum)
 {
   const char name[] = "PE file checksum";
-  if (checksum_show || (checksum_update && changed) || checksum_update > 1)
+  if (checksum_option > 1)
     {
       int fd;
       unsigned new_checksum, old_checksum;
-      if ((fd = open (pathname, (checksum_update ? O_RDWR : O_RDONLY) | O_BINARY)) == -1)
+      if ((fd = open (pathname, (checksum_option > 2 ? O_RDWR : O_RDONLY) | O_BINARY)) == -1)
 	{
 	  fprintf (stderr, "%s: Unable to reopen, errno=%d\n", pathname, errno);
 	  return;
 	}
-      new_checksum = pe32_checksum (fd, checksum_update, &old_checksum);
+      new_checksum = pe32_checksum (fd, checksum_option > 2, &old_checksum);
       close(fd);
       if (!new_checksum)
 	{
@@ -650,19 +675,19 @@ do_checksum (const char *pathname, int indent, int changed, ULONG hdr_checksum)
 	}
 
       if (new_checksum == old_checksum)
-	  printf ("%*s%-24s: 0x%08x (%sverified)\n", indent, "", name,
-		  old_checksum, (checksum_update ? "unchanged, " : ""));
-      else if (checksum_update)
+	  printf ("%*s%-24s: 0x%08x (%schanged, verified)\n", indent, "", name,
+	          old_checksum, (changed ? "" : "un"));
+      else if (checksum_option > 2)
 	  printf ("%*s%-24s: 0x%08x (updated from 0x%08x)\n", indent, "", name,
 		  new_checksum, old_checksum);
       else
 	  printf ("%*s%-24s: 0x%08x (*needs update to 0x%08x*)\n", indent, "", name,
 		  old_checksum, new_checksum);
     }
-  else /* (!checksum_show && checksum_update == 1 && !changed) */
+  else /* (checksum_option == 1) */
     {
-      printf ("%*s%-24s: 0x%08x (unchanged, not verified)\n", indent, "", name,
-	      hdr_checksum);
+      printf ("%*s%-24s: 0x%08x (%schanged, not verified)\n", indent, "", name,
+	      hdr_checksum, (changed ? "" : "un"));
     }
 }
 
@@ -767,11 +792,7 @@ handle_checksum_option (const char *option_name,
 {
   int bool_value;
   if (!option_arg)
-    {
-      checksum_show = 1;
-      if (handle_any_sizeof == DONT_HANDLE)
-	handle_any_sizeof = DO_READ;
-    }
+    checksum_option = 1;
   else
     {
       if (string_to_bool (option_arg, &bool_value) != 0)
@@ -781,9 +802,12 @@ handle_checksum_option (const char *option_name,
           short_usage (stderr);
           exit (1);
         }
-      checksum_update = (bool_value ? 2 : 1);
-      handle_any_sizeof = DO_WRITE;
+      checksum_option = (!bool_value ? 2 : 3);
     }
+  if (checksum_option > 2)
+    handle_any_sizeof = DO_WRITE;
+  else if (handle_any_sizeof == DONT_HANDLE)
+    handle_any_sizeof = DO_READ;
 }
 
 void
@@ -1042,6 +1066,11 @@ pe_open (const char *path, BOOL writing)
   int fd;
   void *map;
   
+  /* Get filesize for checksum update. */
+  struct stat st;
+  if (stat (path, &st))
+    return NULL;
+
   fd = open (path, O_BINARY | (writing ? O_RDWR : O_RDONLY));
   if (fd == -1)
     return NULL;
@@ -1053,6 +1082,7 @@ pe_open (const char *path, BOOL writing)
       return NULL;
     }
   pef.pathname = path;
+  pef.filesize = st.st_size;
   pef.dosheader = (PIMAGE_DOS_HEADER) map;
   pef.ntheader32 = (PIMAGE_NT_HEADERS32)
 		   ((PBYTE) pef.dosheader + pef.dosheader->e_lfanew);
@@ -1233,9 +1263,12 @@ help (FILE *f)
 "                              than 2 GB.\n"
 "  -S, --sepdbg       [BOOL]   Debugging information was removed and stored\n"
 "                              separately in another file.\n"
-"  -k, --checksum     [BOOL]   Displays (no argument), updates after changes\n"
-"                              only (false) or always fixes (true) the file\n"
-"                              checksum in the PE header.\n"
+"  -k, --checksum     [BOOL]   Displays (no argument), verifies (false) or\n"
+"                              always fixes (true) the file checksum in the PE\n"
+"                              header.  By default, the checksum is updated\n"
+"                              without reading the whole file under the\n"
+"                              assumption that the original checksum was\n"
+"                              correct.\n"
 "  -x, --stack-reserve [NUM]   Reserved stack size of the process in bytes.\n"
 "  -X, --stack-commit  [NUM]   Initial commited portion of the process stack\n"
 "                              in bytes.\n"
@@ -1272,7 +1305,7 @@ help (FILE *f)
 "For flag values, to set a value, and display the results symbolic, repeat the\n"
 "option:  --tsaware=true --tsaware -d0 -d\n"
 "The time of last modification of each file is preserved unless the checksum\n"
-"has been updated.\n"
+"has been fixed by option --checksum=true\n"
 "\n", f);
 }
 
